import { Intersection, Object3D, Ray, Raycaster, Vector3 } from 'three';
import { MeshEventDescriptor, MeshEventDispatcher } from '@g.frame/core';
import { ParentEvent } from '@g.frame/core';
export declare enum ActionControllerEventName {
    buttonUp = "buttonUp",
    buttonDown = "buttonDown",
    move = "move",
    click = "click",
    over = "over",
    out = "out"
}
export declare class ActionControllerEvent extends ParentEvent<string> {
    eventName: string;
    data: {
        intersection?: IntersectionExt;
        controllerNumber: number;
        context: ActionController;
        ray: Ray;
    };
    constructor(eventName: string, data: {
        intersection?: IntersectionExt;
        controllerNumber: number;
        context: ActionController;
        ray: Ray;
    });
}
/**
 * Intersection Extended interface saves information about intersected Mesh,
 * during raycasting Object3D with recursion flag
 */
export interface IntersectionExt extends Intersection {
    objectIntersected?: Object3D;
    orderNumber?: number;
}
/**
 * ActionController class provides easy THREE.Mesh events listeners.
 * Use next events for callbacks 'buttonDown', 'move', 'buttonUp', 'over', 'out', 'click'
 */
export declare class ActionController extends MeshEventDispatcher {
    /**
     * @ignore
     */
    __agentConstructor: Function;
    private buttonDownRayDirections;
    /**
     * Constructor of the ActionController
     */
    constructor();
    /**
     * Function, that checks if object has invisible parents in it's parent tree.
     * Also works to check if object isn't attached (isn't child of) to scene.
     * @param object Instance of Object3D to test
     * @returns True in next cases: some of parents are invisible or latest parent isn't Scene.
     * False in next case: all parents are visible and the latest parent is instance of Scene.
     */
    static hasInvisibleParent(object: Object3D): boolean;
    /**
     * Function to check whether was click with the hardcoded precision
     * @param eventName Current event name
     * @param buttonDownRayDirections Array of ray directions, when button down was fired for all controllers
     * @param direction Current direction
     * @param controllerNumber Which controller is in use
     * @returns Was click or not
     */
    static checkClick(eventName: ActionControllerEventName, buttonDownRayDirections: Array<Vector3>, direction: Vector3, controllerNumber: number): boolean;
    /**
     * Function to get array of objects to raycast.
     * @param eventName Selected event name
     * @param events Array of events from [[ActionController]]
     * @param wasClick Information about click. Check function [[ActionController.checkClick]]
     * @returns Filtered array of Object3D instances
     */
    static getObjectsToRaycast(eventName: ActionControllerEventName, events: Array<MeshEventDescriptor>, wasClick: boolean): Array<Object3D>;
    /**
     * Function that is used by ActionController to raycast objects.
     * It raycasts Meshes with non-recursive flag and Groups and simple Object3Ds with recursive.
     * Result is already sorted by distance
     * @param objectsToRaycast Array of Object3D instances. Works well with Object3D, Group and Mesh.
     * @param raycaster Defined raycaster, that is used. Don't forget to set .ray of it.
     * @returns Sorted array of intersections.
     * Every intersections has property of objectIntersected, if recursion was used.
     */
    static raycastObjects(objectsToRaycast: Array<Object3D>, raycaster: Raycaster): Array<IntersectionExt>;
    /**
     * Override of the MeshEventDispatcher.fire function to use
     *      ActionControllerEventName as eventName
     *      ActionControllerEvent as event
     * Set's also internal info to mesh.userData
     * @param eventName
     * @param mesh
     * @param event
     */
    fire(eventName: ActionControllerEventName, mesh?: Object3D, event?: ActionControllerEvent): void;
    /**
     * Override of the MeshEventDispatcher.on function to use only ActionControllerEventName as eventName
     * Adds next arrays to mesh.userData (if mesh is presented):
     * ```typescript
     * mesh.userData.isButtonDown = [];
     * mesh.userData.isOver = [];
     * ```
     * @param eventName
     * @param mesh
     * @param callback1 callback or condition if callback2 is defined
     * @param callback2 callback if condition callback1 is defined
     */
    on(eventName: ActionControllerEventName, mesh?: Object3D, callback1?: Function, callback2?: Function): void;
    /**
     * Override of the MeshEventDispatcher.on function to use only ActionControllerEventName as eventName
     * Adds next arrays to mesh.userData (if mesh is presented):
     * ```typescript
     * mesh.userData.isButtonDown = [];
     * mesh.userData.isOver = [];
     * ```
     * @param eventName
     * @param mesh
     * @param callback1 callback or condition if callback2 is defined
     * @param callback2 callback if condition callback1 is defined
     */
    once(eventName: ActionControllerEventName, mesh?: Object3D, callback1?: Function, callback2?: Function): void;
    /**
     * Override of the MeshEventDispatcher.off function to use only ActionControllerEventName as eventName
     * @param eventName
     * @param mesh
     * @param callback
     */
    off(eventName?: ActionControllerEventName, mesh?: Object3D, callback?: Function): void;
    /**
     * Function to raycast and fire events of the meshes
     * @param eventName Event name to raycast
     * @param raycaster Raycaster with defined ray.
     * You can define raycaster from origin point and quaternion using next code:
     * ```typescript
     * // quaternion: Quaternion and position: Vector3 are defined.
     * const direction = new Vector3(0, 0, -1).applyQuaternion(quaternion);
     * const raycaster = new Raycaster();
     * raycaster.ray.set(position, direction);
     * ```
     * @param controllerNumber Controller number from where event was fired. Use 0 for left, 1 for right.
     * For all another with undefined handness use 0 also.
     */
    protected update(eventName: ActionControllerEventName, raycaster: Raycaster, controllerNumber?: number): void;
}
